// Code specific to the client in the client-server mode.

use rand_xoshiro::{rand_core::SeedableRng, Xoshiro256PlusPlus};
use serde::Deserialize;
use std::env;
use std::result::Result;
use tokio::io::AsyncWriteExt;
use tokio::task;
use tokio::time::Duration;
use tokio_socks::tcp::Socks5Stream;

mod messenger;

use crate::messenger::dists::{ConfigDistributions, Distributions};
use crate::messenger::error::MessengerError;
use crate::messenger::state::{
    manage_single_active_conversation, manage_single_idle_conversation, StateMachine,
};

async fn manage_conversation(config: Config) -> Result<(), MessengerError> {
    let mut rng = Xoshiro256PlusPlus::from_entropy();
    let distributions: Distributions = config.distributions.try_into()?;

    struct StrParams<'a> {
        socks: &'a str,
        server: &'a str,
        sender: &'a str,
        group: &'a str,
    }
    let str_params = StrParams {
        socks: &config.socks,
        server: &config.server,
        sender: &config.sender,
        group: &config.group,
    };

    let mut state_machine = StateMachine::start(distributions, &mut rng);
    let recipients: Vec<&str> = config.recipients.iter().map(String::as_str).collect();

    async fn error_collector<'a>(
        bootstrap: Option<Duration>,
        str_params: &'a StrParams<'a>,
        rng: &mut Xoshiro256PlusPlus,
        mut state_machine: StateMachine,
        recipients: Vec<&str>,
    ) -> Result<(), (StateMachine, MessengerError)> {
        let mut stream = match Socks5Stream::connect_with_password(
            str_params.socks,
            str_params.server,
            str_params.sender,
            str_params.group,
        )
        .await
        {
            Ok(stream) => stream,
            Err(e) => return Err((state_machine, e.into())),
        };

        if let Err(e) = stream
            .write_all(&mgen::serialize_str(str_params.sender))
            .await
        {
            return Err((state_machine, e.into()));
        }

        if let Some(bootstrap) = bootstrap {
            tokio::time::sleep(bootstrap).await;
        }

        loop {
            state_machine = match state_machine {
                StateMachine::Idle(conversation) => {
                    manage_single_idle_conversation(
                        conversation,
                        &mut stream,
                        str_params.sender,
                        recipients.clone(),
                        rng,
                    )
                    .await?
                }
                StateMachine::Active(conversation) => {
                    manage_single_active_conversation(
                        conversation,
                        &mut stream,
                        str_params.sender,
                        recipients.clone(),
                        rng,
                    )
                    .await?
                }
            };
        }
    }

    let retry = config.retry;
    let retry = Duration::from_secs_f64(retry);

    match error_collector(
        Some(Duration::from_secs_f64(config.bootstrap)),
        &str_params,
        &mut rng,
        state_machine,
        recipients.clone(),
    )
    .await
    .expect_err("Inner loop returned Ok?")
    {
        (sm, MessengerError::Recoverable(_)) => {
            state_machine = sm;
            tokio::time::sleep(retry).await;
        }
        (_, e) => return Err(e),
    };

    loop {
        match error_collector(
            None,
            &str_params,
            &mut rng,
            state_machine,
            recipients.clone(),
        )
        .await
        .expect_err("Inner loop returned Ok?")
        {
            (sm, MessengerError::Recoverable(_)) => {
                state_machine = sm;
                tokio::time::sleep(retry).await;
            }
            (_, e) => return Err(e),
        };
    }
}

#[derive(Debug, Deserialize)]
struct Config {
    sender: String,
    group: String,
    recipients: Vec<String>,
    socks: String,
    server: String,
    bootstrap: f64,
    retry: f64,
    distributions: ConfigDistributions,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let mut args = env::args();
    let _ = args.next();
    let mut handles = vec![];
    for config_file in args {
        let toml_s = std::fs::read_to_string(config_file)?;
        let config = toml::from_str(&toml_s)?;
        let handle: task::JoinHandle<Result<(), MessengerError>> =
            tokio::spawn(manage_conversation(config));
        handles.push(handle);
    }
    for handle in handles {
        handle.await??;
    }
    Ok(())
}
