use mgen::{log, parse_identifier, SerializedMessage};
use std::collections::HashMap;
use std::error::Error;
use std::result::Result;
use std::sync::{Arc, Mutex};
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::net::{
    tcp::{OwnedReadHalf, OwnedWriteHalf},
    TcpListener,
};
use tokio::sync::{mpsc, watch};

// FIXME: identifiers should be interned
type ID = String;

#[tokio::main]
async fn main() -> Result<(), Box<dyn Error>> {
    let listener = TcpListener::bind("127.0.0.1:6397").await?;

    log!("Listening");

    // FIXME: should probably be a readers-writer lock
    let snd_db = Arc::new(Mutex::new(HashMap::<
        ID,
        mpsc::UnboundedSender<Arc<SerializedMessage>>,
    >::new()));
    let mut writer_db = HashMap::<ID, mpsc::Sender<(OwnedWriteHalf, watch::Sender<bool>)>>::new();

    loop {
        let (socket, _) = listener.accept().await?;
        let (mut rd, wr) = socket.into_split();

        let id = parse_identifier(&mut rd).await?;
        log!("Accepting \"{id}\"");
        if let Some(socket_updater) = writer_db.get(&id) {
            // we've seen this client before

            // start the new reader thread with a new watch
            let (watch_snd, watch_rcv) = watch::channel(false);
            spawn_message_receiver(rd, snd_db.clone(), watch_rcv);

            // give the writer thread the new write half of the socket and watch
            socket_updater.send((wr, watch_snd)).await?;
        } else {
            // newly-registered client
            log!("New client");

            // message channel, used to send messages between threads
            let (msg_snd, msg_rcv) = mpsc::unbounded_channel::<Arc<SerializedMessage>>();

            let snd_db = snd_db.clone();
            {
                let mut locked_db = snd_db.lock().unwrap();
                locked_db.insert(id.clone(), msg_snd);
            }

            // socket watch, used to terminate the socket if the sender encounters an error
            let (watch_snd, watch_rcv) = watch::channel(false);

            // socket updater, used to give the sender thread a new socket and watch
            let (socket_updater_snd, socket_updater_rcv) = mpsc::channel(8);
            socket_updater_snd.send((wr, watch_snd)).await?;

            writer_db.insert(id.clone(), socket_updater_snd);

            spawn_message_receiver(rd, snd_db, watch_rcv);
            tokio::spawn(async move {
                send_messages(msg_rcv, socket_updater_rcv).await;
            });
        }
    }
}

fn spawn_message_receiver(
    rd: OwnedReadHalf,
    db: Arc<Mutex<HashMap<ID, mpsc::UnboundedSender<Arc<SerializedMessage>>>>>,
    mut watch_rcv: watch::Receiver<bool>,
) {
    watch_rcv.borrow_and_update();
    tokio::spawn(async move {
        tokio::select! {
            ret = get_messages(rd, db) => {
                if let Err(e) = ret {
                    log!("message receiver failed: {:?}", e);
                }
            }
            _ = watch_rcv.changed() => {
                log!("receiver terminated");
                // should cause the other thread to terminate, dropping the socket
            }
        }
    });
}

/// Loop for receiving messages on the socket, figuring out who to deliver them to,
/// and forwarding them locally to the respective channel.
async fn get_messages(
    mut socket: OwnedReadHalf,
    db: Arc<Mutex<HashMap<ID, mpsc::UnboundedSender<Arc<SerializedMessage>>>>>,
) -> Result<(), Box<dyn Error>> {
    // stores snd's for contacts this client has already sent messages to, to reduce contention on the main db
    // if memory ends up being more of a constraint, could be worth getting rid of this
    let mut localdb = HashMap::<ID, mpsc::UnboundedSender<Arc<SerializedMessage>>>::new();
    let mut sender = "".to_string();
    loop {
        log!("waiting for message from {}", sender);
        let (message, buf) = mgen::get_message(&mut socket).await?;
        log!(
            "got message from {} for {:?}",
            message.sender,
            message.recipients
        );
        sender = message.sender;
        let missing = message
            .recipients
            .iter()
            .filter(|&r| !localdb.contains_key(r))
            .collect::<Vec<&ID>>();

        {
            let locked_db = db.lock().unwrap();
            for m in missing {
                if let Some(snd) = locked_db.get(m) {
                    localdb.insert(m.to_string(), snd.clone());
                }
            }
        }

        let m = Arc::new(SerializedMessage {
            header: buf,
            body: message.body,
        });

        for recipient in message.recipients.iter() {
            let recipient_sender = localdb
                .get(recipient)
                .unwrap_or_else(|| panic!("Unknown sender: {}", recipient));
            recipient_sender
                .send(m.clone())
                .expect("Recipient closed channel with messages still being sent");
        }
    }
}

/// Loop for receiving messages on the mpsc channel for this recipient,
/// and sending them out on the associated socket.
async fn send_messages(
    mut msg_rcv: mpsc::UnboundedReceiver<Arc<SerializedMessage>>,
    mut socket_updater: mpsc::Receiver<(OwnedWriteHalf, watch::Sender<bool>)>,
) {
    let (mut current_socket, mut current_watch) =
        socket_updater.recv().await.expect("socket updater closed");
    let mut message_cache = None;
    loop {
        let message = if message_cache.is_none() {
            msg_rcv.recv().await.expect("message channel closed")
        } else {
            message_cache.unwrap()
        };
        log!("sending message");
        if message.write_all_to(&mut current_socket).await.is_err()
            || current_socket.flush().await.is_err()
        {
            message_cache = Some(message);
            log!("terminating connection");
            // socket is presumably closed, clean up and notify the listening end to close
            // (all best-effort, we can ignore errors because it presumably means it's done)
            let _ = current_watch.send(true);
            let _ = current_socket.shutdown().await;

            // wait for the new socket
            (current_socket, current_watch) =
                socket_updater.recv().await.expect("socket updater closed");
            log!("socket updated");
        } else {
            message_cache = None;
        }
    }
}
